/*

  File LUtrid.cpp:

  contiene la definizione delle funzioni membro della classe LUtrid, che
  definisce una matrice tridiagonale 3xn, la fattorizza LU e risolve un 
  sistema di dimensione n con la stessa matrice.

*/


#include <iostream>
#include "LUtrid.h"
#include <stdlib.h>
using namespace std;


/*

 LUtrid(): costruttore della classe

*/

LUtrid::LUtrid( int N )
{
  dim = N;
  mat = new double * [3];
  for ( int i=0; i < 3; i++ ) mat[i] = new double[N];
  tmp = new double[N];           // Vettore per memorizzazione temporanea dei secondi membri
}


/*

  ~LUtrid(): distruttore della classe.

*/

LUtrid::~LUtrid()
{
  for ( int i = 0; i < 3; i++ ) delete( mat[i] );
  delete( mat );
  delete( tmp );
}



/*

  luFact( ): effettua la fattorizzazione LU di una matrice a[3][dim]
  secondo l'algoritmo di Thomas

*/


void LUtrid::luFact( )
{

  int i;

  for ( i = 1; i < dim; i++ )
    {
      mat[0][i] = mat[0][i] / mat[1][i-1];
      mat[1][i] = mat[1][i] - mat[0][i] * mat[2][i-1];
    }

}


/*

  luSolve( x, b ): esegue l'inversione di un sistema tridiagonale a*x=b
  secondo l'algoritmo di Thomas.

*/


void LUtrid::luSolve( double * x, double * b )
{

  int i;

  // Forward substitution

  x[0] = b[0];    // Primo elemento
  for ( i = 1; i < dim; i++ )
    {
      x[i] = b[i] - mat[0][i] * x[i-1];
    }

  // Ricopio il vettore soluzione della FS in f

  for ( i = 0; i < dim; i++ ) tmp[i] = x[i];

  // Backward substitution

  x[dim-1] = tmp[dim-1] / mat[1][dim-1];
  for ( i = dim - 2; i >= 0; i-- )
    {
      x[i] = ( tmp[i] - mat[2][i] * x[i+1] ) / mat[1][i];
    }

}


/*

   Funzione setDiagonal:
   definisce la matrice a partire dalle diagonali b, a, c.
   Nota che i tre vettori debbono rispettare la condizione:
   b[0] = 0.0; c[dim-1] = 0.0.

*/

void LUtrid::setDiagonal( double b[], double a[], double c[] )
{
  if ( b[0] != 0.0 )
    {
      cout << "Errore: l\'elemento b[0] deve essere nullo" << endl;
      exit( 0 );
    }

  if ( c[dim-1] != 0.0 )
    {
      cout << "Errore: l\'elemento c[N-1] deve essere nullo" << endl;
      exit( 0 );
    }

  for ( int j = 0; j < dim; j++ ) mat[0][j] = b[j];
  for ( int j = 0; j < dim; j++ ) mat[1][j] = a[j];
  for ( int j = 0; j < dim; j++ ) mat[2][j] = c[j];
}


/*

  Funzione setElement:
  inizializza l'elemento a_ij della matrice al valore passato.
  Nota che in questo caso l'elemento si riferisce a quello della matrice
  originale, NON quella memorizzata per diagonali!

*/

void LUtrid::setElement( int i, int j, double a_ij )
{
  mat[0][0] = 0.0;                   // Azzera per sicurezza il valore b_1
  mat[2][dim-1] = 0.0;               // Azzera per sicurezza il valore c_N
  if ( ( ( i >= 0 ) && ( i < dim ) ) && ( ( j >= 0 ) && ( j < dim ) ) )
    {
      if ( i-1 == j ) mat[0][i] = a_ij;
      if ( i == j )   mat[1][i] = a_ij;
      if ( i+1 == j ) mat[2][i] = a_ij;
    }
}


/*

  Funzione getElement:
  restituisce il valore dell'elemento i,j della matrice.
  Nota che in questo caso l'elemento si riferisce a quello della matrice
  originale, NON quella memorizzata per diagonali!

*/

double LUtrid::getElement( int i, int j )
{
  double elem;

  if ( ( ( i >= 0 ) && ( i < dim ) ) && ( ( j >= 0 ) && ( j < dim ) ) )
    {
      elem = 0.0;
      if ( i-1 == j ) elem = mat[0][i];
      if ( i == j )   elem = mat[1][i];
      if ( i+1 == j ) elem = mat[2][i];
    }
  return( elem );
}
