/*

  File Problema_ai_limiti.ccp:
  Risolve l'equazione: y'' = -4*sin(2*x), x in [0,pi/4]
  con condizioni al bordo: y(0)=0.0, y(pi/4)=1.0 se di tipo Dirichlet
  oppure y'(0)=2.0, y(pi/4)=1.0 se di tipo Neumann al bordo sinistro.
  Nota che non si possono imporre condizioni di tipo Neumann a entrambi
  i bordi, altrimenti la soluzione risulta indeterminata a meno di una 
  costante!
  La soluzione dell'equazione e': y(x)=sin(2*x)

*/


#include <iostream>
#include <fstream>
#include <cmath>
#include "LUtrid.h"

using namespace std;


main()
{
  cout << "Calcolo della soluzione dell\'equazione differenziale:" << endl;
  cout << "d^2 y / dx^2 = -4*sin(2*x)" << endl;

  // Input del numero di punti griglia
  long int N;
  cout << "Inserisci il numero di punti da utilizzare: ";
  cin >> N;

  // Condizioni al bordo
  int tcb;
  cout << "Condizioni al bordo di tipo Dirichlet (0) o Neumann (1) al bordo sinistro? ";
  cin >> tcb;


  // Dimensione del problema, dipende da condizioni al bordo:
  // dim = N-2 per condizioni Dirichlet
  // dim = N-1 per condizioni di Neumann al bordo sinistro

  long int dim;
  if ( tcb == 0 )
    dim = N - 2;     // Dirichlet
  else
    dim = N-1;       // Neumann

  // Definizione dell'intervallo di definizione di x

  double *x = new double[N];
  double pi = acos( -1.0 );
  double a = 0.0, b = pi / 4.0;
  double dx = ( b - a ) / ( N - 1 );

  for ( int i = 0; i < N; i++ ) x[i] = a + i * dx;


  // Definizione dei vettori delle diagonali della  matrice della derivata seconda

  double *bb = new double[dim];
  double *aa = new double[dim];
  double *cc = new double[dim];
  double dx2 = dx * dx;

  bb[0] = 0.0;
  aa[0] = - 2.0;
  if ( tcb == 0 )
    cc[0] = 1.0;            // Dirichlet
  else
    cc[0] = 2.0;            // Neumann

  for ( int i = 1; i < dim - 1; i++ )
    {
      bb[i] = 1.0;
      aa[i] = - 2.0;
      cc[i] = 1.0;
    }

  bb[dim-1] = 1.0;          // Neumann (solo al bordo sinistro, 
                            // quindi a destra coincide con Dirichlet!)
  aa[dim-1] = - 2.0;
  cc[dim-1] = 0.0;


  // Definizione delle condizioni al bordo

  double ya;
  if ( tcb == 0 )
    ya = 0.0;         // Dirichlet
  else
    ya = 2.0;         // Neumann

  double yb = 1.0;         // Dirichlet & Neumann!


  // Creazione dell'oggetto LUtrid e fattorizzazione LU della matrice

  LUtrid d2 = LUtrid( dim );
  d2.setDiagonal( bb, aa, cc );
  d2.luFact( );


  // Costruisce i secondi membri dell'equazione

  double *sm = new double[dim];

  for ( int i = 0; i < dim; i++ )
    {
      if ( tcb == 0 )
	sm[i] = - 4.0 * sin( 2.0 * x[i+1] ) * dx2;   // Dirichlet
      else
	sm[i] = - 4.0 * sin( 2.0 * x[i] ) * dx2;     // Neumann
    }
  if ( tcb == 0 )
    sm[0]     = sm[0] - ya;                 // Dirichlet
  else
    sm[0]     = sm[0] + 2.0 * dx * ya;      // Neumann

  sm[dim-1] = sm[dim-1] - yb;             // Dirichlet & Neumann!


  // Inverte il sistema fattorizzato LU

  double *y = new double[dim];

  d2.luSolve( y, sm );


  // Stampa del risultato

  char fname[250];
  double yex, err;

  cout << "Nome del file dei risultati: ";
  cin >> fname;

  ofstream outFile;
  outFile.open( fname );
  outFile.precision(15);
  outFile << "# x" << "  " << "y" << "  " << "y_ex" << "  " << "errore" << endl;
  // Riga presente SOLO se si hanno condizioni al bordo di tipo Dirichlet!
  if ( tcb == 0 ) outFile << x[0] << "  " << ya << "  " << ya << "  " << 0.0 << endl;
  for ( int i = 0; i < dim; i++ )
    {
      if ( tcb == 0 )
	yex = sin( 2.0 * x[i+1] );      // Dirichlet
      else
	yex = sin( 2.0 * x[i] );        // Neumann
      err = abs( yex - y[i] );
      if ( tcb == 0 )
	outFile << x[i+1] << "  " << y[i] << "  " << yex << "  " << err << endl;   // Dirichlet
      else
	outFile << x[i] << "  " << y[i] << "  " << yex << "  " << err << endl;     // Neumann
    }
  outFile << x[N-1] << "  " << yb << "  " << yb << "  " << 0.0 << endl;
  outFile.close();

}
